import asyncio

from airflow.notifications.basenotifier import BaseNotifier
from telegram import Bot

from dump.config_utils import load_config


class TelegramNotification(BaseNotifier):
    def __init__(
        self,
        tg_notification_config_name: str = "tg_notification.ini",
        tg_notification_config_section: str = "default",
    ):
        self.__config = load_config(
            filename=tg_notification_config_name,
            section=tg_notification_config_section,
        )

    @staticmethod
    async def send_tg_notifier(
        notifier_message,
        telegram_bot_token,
        telegram_chat_id,
    ):

        bot = Bot(token=telegram_bot_token)
        await bot.sendMessage(
            chat_id=telegram_chat_id,
            text=notifier_message,
            parse_mode="Markdown",
        )

    def notify(self, context):
        telegram_bot_token, telegram_chat_id, airflow_url = self.__config.values()

        task_id = context["ti"].task_id.replace("_", "\_")
        task_state = context["ti"].state.replace("_", "\_")
        task_log_url = (
            context["ti"]
            .log_url.replace("http://localhost:8080", airflow_url)
            .replace("_", "\_")
        )
        dag_name = context["ti"].dag_id.replace("_", "\_")

        message_template = (
            f"***Dag name:*** {dag_name} \n"
            f"***Task id:*** {task_id} \n"
            f"***Task State:*** \N{Cross mark}{task_state}\N{Cross mark} \n"
            f"***Task Log URL:*** {task_log_url} \n"
        )

        asyncio.run(
            self.send_tg_notifier(
                notifier_message=message_template,
                telegram_bot_token=telegram_bot_token,
                telegram_chat_id=telegram_chat_id,
            )
        )



import datetime
from typing import Dict, Optional

import dateutil
import numpy as np
import pandas as pd
from pandas import DataFrame as PandasDataFrame


BOOL_MAPPER = {"f": False, "t": True}


def cast_df_by_schema(
    df: PandasDataFrame, schema: Dict[str, str], cast_timestamp: bool = True
) -> PandasDataFrame:
    for column in df.columns:
        if column in schema:
            cast_to = schema[column]["data_type"]
            try:
                if cast_to == "timestamp":
                    if cast_timestamp:
                        df[column] = pd.to_datetime(
                            df[column], errors="coerce", format="mixed"
                        )
                    continue

                if cast_to == "bool":
                    df[column] = df[column].apply(lambda x: BOOL_MAPPER.get(x, np.nan))

                df = df.astype({column: cast_to})
            except Exception as e:
                print(f"unable to cast column: {column} to {cast_to} error: {e}")
                pass
    df = df.replace("nan", np.nan)
    df = df.replace("None", np.nan)
    return df



from typing import Any, Dict

import clickhouse_connect
from pandas import DataFrame as PandasDataFrame

from dump.config_utils import load_config


CH_TO_PYTHON_TYPE_MAPPER = {
    # Boolean
    "Bool": "bool",
    # Integers (signed)
    "Int8": "Int8",
    "Int16": "Int16",
    "Int32": "Int32",
    "Int64": "Int64",
    # Integers (unsigned)
    "UInt8": "Int8",
    "UInt16": "Int16",
    "UInt32": "Int32",
    "UInt64": "Int64",
    # Floating point
    "Float32": "float32",
    "Float64": "float64",
    # Temporal
    "timestamp": "timestamp",
    "date": "timestamp",
    # Strings
    "String": "str",
}


class ConnectorCH:
    def __init__(
        self,
        db_config_name: str = "click_house",
    ) -> None:
        self.db_config_name = db_config_name
        self.__config = load_config(section=self.db_config_name)

        self._client = None

    @property
    def client(self):
        if self._client is None:
            try:
                self._client = clickhouse_connect.get_client(**self.__config)
                # print("Connected to the ClickHouse server.")
            except Exception as error:
                print(error)
        return self._client


class TableCH(ConnectorCH):
    def __init__(self, db_config_name: str = "click_house") -> None:
        super().__init__(db_config_name)

    def get_df(self, query: str) -> PandasDataFrame:
        return self.client.query_df(query)


class DBUtilsCH(ConnectorCH):
    @staticmethod
    def _validate_output(inp: list):
        return [x[0] for x in inp]

    def get_table_structure(
        self, table_name: str, database: str = "default"
    ) -> Dict[str, Any]:
        query = f"""
            SELECT name, type, position
            FROM system.columns
            WHERE table = '{table_name}' AND database = '{database}'
        """

        schema = {}
        result = self.client.query(query)
        for column_name, data_type, position in result.result_rows:
            orig_data_type = data_type
            if "int" in data_type and data_type not in CH_TO_PYTHON_TYPE_MAPPER.keys():
                data_type = "integer"
            if (
                "numeric" in data_type or "decimal" in data_type
            ) and data_type not in CH_TO_PYTHON_TYPE_MAPPER.keys():
                data_type = "decimal"
            if "time" in data_type:
                data_type = "timestamp"

            schema[column_name] = {
                "data_type": CH_TO_PYTHON_TYPE_MAPPER.get(data_type, "str"),
                "orig_data_type": orig_data_type,
                "is_nullable": False,
                "position": position,
            }
        return schema



import logging
import os
import weakref
from tempfile import TemporaryDirectory


class TemporaryFileSystem:
    """
    Temporary directory utils class
    used for data manipulators
    """

    def __init__(self, file_format: str = "csv") -> None:
        self.file_format = file_format
        self._tmp_dir = TemporaryDirectory()
        self._last_path: str
        self._finalizer = weakref.finalize(self, self._safe_cleanup, self._tmp_dir.name)

    def __del__(self) -> None:
        """Fallback cleanup (not guaranteed but added as safety net)."""
        self.cleanup()

    @staticmethod
    def _safe_cleanup(dir_path: str) -> None:
        """Safely remove directory ignoring some errors."""
        try:
            if os.path.exists(dir_path):
                for root, dirs, files in os.walk(dir_path, topdown=False):
                    for name in files:
                        os.unlink(os.path.join(root, name))
                    for name in dirs:
                        os.rmdir(os.path.join(root, name))
                os.rmdir(dir_path)
        except (OSError, PermissionError):
            pass

    def cleanup(self) -> None:
        """Explicit cleanup method. Safe to call multiple times."""
        if hasattr(self, "_finalizer") and self._finalizer.alive:
            self._finalizer()

    @property
    def tmp_dir(self) -> str:
        return self._tmp_dir.name

    @property
    def files(self) -> list:
        return os.listdir(str(self.tmp_dir))

    @property
    def next_filename(self) -> str:
        number = 0
        avaliable = [int(x.split(".")[0].split("_")[-1]) for x in self.files]
        if len(avaliable) != 0:
            number = max(avaliable) + 1
        return f"data_{number}.{self.file_format}"

    def save_path(self) -> str:
        self._last_path = os.path.join(str(self.tmp_dir), self.next_filename)
        return self._last_path

    @property
    def dir_info(self) -> None:
        """
        Shows temporary directory info
        file - file size in mb
        """
        for file in self.files:
            check_file = os.path.join(self.tmp_dir, file)
            size = os.path.getsize(check_file) / 1024**2
            print(f"{check_file}: {size} mb")



import logging
import os
import warnings
from typing import Dict, Optional

import pyarrow as pa
import pyarrow.orc as orc
from pandas import DataFrame as PandasDataFrame

from dump.files.tmp_util import TemporaryFileSystem
from dump.postgres.types_util import cast_df_by_schema
from dump.s3.connectors import S3Upload


TABLE_SCHEMA_TO_ORC_MAPPER = {
    # Integer types
    "Int8": pa.int8(),
    "Int16": pa.int16(),
    "Int32": pa.int32(),
    "Int64": pa.int64(),
    # "UInt8": pa.uint8(),
    # "UInt16": pa.uint16(),
    # "UInt32": pa.uint32(),
    # "UInt64": pa.uint64(),
    # Float types
    "float16": pa.float16(),
    "float32": pa.float32(),
    "float64": pa.float64(),
    "float": pa.float64(),
    # Text types
    "str": pa.string(),
    "string": pa.string(),
    "object": pa.string(),
    # Boolean
    "bool": pa.bool_(),
    "boolean": pa.bool_(),
    # Temporal types
    "timestamp": pa.timestamp("ns", tz="+00:00"),
    # "timestamp": pa.timestamp('ns'),
    # "timestamp[ns]": pa.timestamp('ns'),
    # "timestamp[us]": pa.timestamp('us'),
    # "datetime64[ns]": pa.timestamp('ns'),
    # "date": pa.date32(),
    # "timedelta": pa.duration('ns'),
}


class S3UploadDFORC(S3Upload):
    def __init__(
        self,
        bucket_name,
        object_folder="",
        compession_engine: str = "ZLIB",
        compression_strategy: str = "COMPRESSION",
        creds_section="test",
    ):
        super().__init__(bucket_name, object_folder, creds_section)
        self.compession_engine = compession_engine
        self.compression_strategy = compression_strategy

        # for backward compatibility
        self.df_compression_format = "orc"
        self._tmp_file_system = TemporaryFileSystem(file_format="orc")

    @staticmethod
    def __get_columns_right_order(table_schema: Dict[str, str]) -> list:
        """
        Return list of columns sorted by order in destination table
        """
        column_positions = {
            column: meta_info["position"] for column, meta_info in table_schema.items()
        }
        columns_sorted_by_position = sorted(column_positions, key=column_positions.get)
        return columns_sorted_by_position

    def cast_df_by_schema(
        self, df: PandasDataFrame, table_schema: Dict[str, Dict[str, str]]
    ) -> pa.Table:
        df_cols = set(df.columns)
        schema_cols = set(table_schema)
        for del_col in schema_cols - df_cols:
            logging.info(
                f"removing column {del_col} from table_schema, cause its undefiend in df."
            )
            table_schema.pop(del_col)
        df = cast_df_by_schema(df, table_schema, cast_timestamp=True)

        arrow_schema = {}
        for col, setings in table_schema.items():
            arrow_schema[col] = TABLE_SCHEMA_TO_ORC_MAPPER.get(
                setings["data_type"], pa.string()
            )
        arrow_schema = pa.schema(arrow_schema.items())
        df_arrow = pa.Table.from_pandas(df, schema=arrow_schema)

        columns_right_order = self.__get_columns_right_order(table_schema)
        df_arrow = df_arrow.select(columns_right_order)
        return df_arrow

    def upload_to_s3_df(
        self,
        df: PandasDataFrame,
        file_name: str,
        table_schema: Optional[Dict[str, Dict[str, str]]] = None,
    ) -> PandasDataFrame:
        if table_schema:
            df_arrow = self.cast_df_by_schema(df, table_schema)
        else:
            df_arrow = pa.Table.from_pandas(df)

        save_path = self._tmp_file_system.save_path()
        orc.write_table(
            df_arrow,
            save_path,
            compression=self.compession_engine,
            compression_strategy=self.compression_strategy,
        )
        logging.info(f"orc DF meta: {df_arrow}")
        logging.info(f"uploading to s3, object folder: {self.object_folder}")
        logging.info(f"uploading to s3, filename: {file_name}")
        self.upload_to_s3(save_path, file_name)



import psycopg2
from psycopg2 import connect

from dump.config_utils import load_config


TYPE_MAPPER = {
    # Integer types
    "smallint": "Int16",
    "int2": "Int16",
    "integer": "Int32",
    "int": "Int32",
    "int4": "Int32",
    "bigint": "Int64",
    "int8": "Int64",
    "smallserial": "Int16",
    "serial2": "Int16",
    "serial": "Int32",
    "serial4": "Int32",
    "bigserial": "Int64",
    "serial8": "Int64",
    # Float types
    "real": "float32",
    "float4": "float32",
    "double precision": "float64",
    "float8": "float64",
    "decimal": "float64",
    "numeric": "float64",
    # Other types
    "USER-DEFINED": "str",
    "character varying": "str",
    "character": "str",
    "text": "str",
    "boolean": "bool",
    "timestamp without time zone": "timestamp",
    "timestamp": "timestamp",
    "date": "timestamp",
}


class ConnectorPG:
    def __init__(
        self,
        db_config_name: str = "postgres",
    ) -> None:
        self.db_config_name = db_config_name
        self.__config = load_config(section=self.db_config_name)

        self._conn: connect = None
        self._cursor = None

    @property
    def conn(self):
        if self._conn is None:
            try:
                # connecting to the PostgreSQL server
                with psycopg2.connect(**self.__config) as pgconn:
                    # print("Connected to the PostgreSQL server.")
                    self._conn = pgconn
            except (psycopg2.DatabaseError, Exception) as error:
                raise error
        return self._conn

    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.conn.cursor()
        return self._cursor

    @property
    def close(self):
        self._conn.close()

        self._conn = None
        self._cursor = None

    @property
    def commit(self):
        self.conn.commit()

    def execute(self, query: str):
        self.cursor.execute(query)
        self.conn.commit()

    def fetchall(self, query: str) -> list:
        self.execute(query)
        values = self.cursor.fetchall()
        self.conn.commit()

        return values


class DBUtilsPG(ConnectorPG):
    @staticmethod
    def _validate_output(inp: list):
        return [x[0] for x in inp]

    @property
    def _schemas(self) -> list:
        query = """
            SELECT schema_name
            FROM information_schema.schemata;
        """
        return self._validate_output(self.fetchall(query))

    def get_tables_in_schema(self, schema: str) -> list:
        query = f"""
            SELECT table_name
            FROM information_schema.tables
            WHERE table_schema = '{schema}'
        """
        tables = self._validate_output(self.fetchall(query))
        return tables

    def get_table_structure(self, table_name: str, db_schema: str) -> dict:
        query = f"""
            SELECT column_name, data_type, is_nullable, ordinal_position
            FROM information_schema.columns
            WHERE table_name = '{table_name}' and
            table_schema = '{db_schema}';
        """

        schema = {}
        for column_name, data_type, is_nullable, position in self.fetchall(query):
            orig_data_type = data_type
            if "int" in data_type and data_type not in TYPE_MAPPER.keys():
                data_type = "integer"
            if (
                "numeric" in data_type or "decimal" in data_type
            ) and data_type not in TYPE_MAPPER.keys():
                data_type = "decimal"
            if "time" in data_type:
                data_type = "timestamp"

            schema[column_name] = {
                "data_type": TYPE_MAPPER.get(data_type, "str"),
                "orig_data_type": orig_data_type,
                "is_nullable": is_nullable,
                "position": position,
            }
        return schema

    def get_schema_by_table(self, table_name: str) -> str:
        query = f"""
            SELECT table_schema
            FROM information_schema.tables
            WHERE table_name = '{table_name}';
        """
        schema = self._validate_output(self.fetchall(query))
        return schema[0]

    def get_primary_keys(self, table_name: str, db_schema: str) -> list:
        query = f"""
            select column_name
            from information_schema.key_column_usage
            where table_schema = '{db_schema}'
            and table_name = '{table_name}'
        """
        keys = self._validate_output(self.fetchall(query))
        return keys


class SequenceUtils:
    def __init__(
        self,
        sequence_name: str = "seq_general_job_id",
        sequence_schema: str = "public",
        db_config_name: str = "postgres",
    ) -> None:
        self.sequence_name = sequence_name
        self.sequence_schema = sequence_schema

        self.connector = ConnectorPG(db_config_name)

    @property
    def next_sequence(self):
        query = f"""
            SELECT nextval('{self.sequence_schema}.{self.sequence_name}');
        """
        sequence = self.connector.fetchall(query)[0][0]
        return sequence

    @property
    def current_sequence(self):
        query = f"""
            SELECT last_value FROM {self.sequence_schema}.{self.sequence_name}
        """
        sequence = self.connector.fetchall(query)[0][0]
        return sequence


# class TemporaryTablePG(ConnectorPG):
#     def __init__(self, name: str, db_config_name: str = "postgres") -> None:
#         super().__init__(db_config_name)
#         self.name = name
#         self.data_manipulator = DataManupulationPG(db_config_name=self.db_config_name)

#         # mapping dtypes to PG types
#         self._KIND_MAPPER = {
#             "M": "timestamp",
#             "f": "real",
#             "i": "bigint",
#         }

#     def _pd_types_to_pg_types(self, df: PandasDataFrame) -> dict:
#         df_types = df.dtypes.to_dict()
#         some = {
#             col: self._KIND_MAPPER.get(col_type.kind, "varchar")
#             for col, col_type in df_types.items()
#         }
#         return some

#     def _types_to_str_statments(self, types: dict) -> str:
#         types_statments = ", \n ".join([" ".join(x) for x in types.items()])
#         return types_statments

#     def create_table(self, df: PandasDataFrame):
#         types = self._pd_types_to_pg_types(df)
#         types_statments = self._types_to_str_statments(types)
#         query = f"""
#         CREATE TEMP TABLE {self.name}(
#             {types_statments}
#         );
#         """
#         self.execute(query)
#         self.commit



import csv
from typing import Dict, Optional

import pandas as pd
from pandas import DataFrame as PandasDataFrame

from dump.files.tmp_util import TemporaryFileSystem

from .connectors import ConnectorMySQL


class GetDataManipulatorMySQL(ConnectorMySQL):
    def __init__(
        self,
        encoding="utf-8",
        size: Optional[int] = 8388608,
        db_config_name: str = "postgres",
    ) -> None:
        super().__init__(db_config_name=db_config_name)
        self.encoding = encoding
        self.size = size

        # requires for not execute same query more than one time
        self.__query_history: Dict[str, str] = {}
        self._tmp_file_system = TemporaryFileSystem()

    def clear_query_history(self):
        self.__query_history: Dict[str, str] = {}

    def get_data(self, query: str, path: Optional[str] = None) -> str:
        """
        fetching data from mysql/mariaDB using query
        and saving into csv locating in temporary directory
        """
        if path is None:
            path = self._tmp_file_system.save_path()

        if query not in self.__query_history:
            results = self.fetchall(query)
            with open(path, "w", encoding=self.encoding) as csvfile:
                csvwriter = csv.writer(csvfile)
                csvwriter.writerow(
                    [i[0] for i in self.cursor.description]
                )  # Write headers
                csvwriter.writerows(results)  # Write data

                self.__query_history[query] = self._tmp_file_system._last_path

        path = self.__query_history[query]
        self.close
        return path

    def get_df(self, query: str) -> PandasDataFrame:
        path = self.get_data(query)
        return pd.read_csv(path, encoding=self.encoding, low_memory=False)



import warnings
from typing import Optional

from pandas import DataFrame as PandasDataFrame

from dump.postgres.types_util import cast_df_by_schema

from .connectors import ConnectorMySQL, DBUtilsMySQL
from .manipulators import GetDataManipulatorMySQL


class TableMySQL(ConnectorMySQL):
    def __init__(
        self,
        table_name: str,
        db_schema: Optional[str] = None,
        limit: int = None,
        auto_cast: Optional[bool] = False,
        clear_tech_cols: Optional[bool] = False,
        size: Optional[int] = 8388608,
        db_config_name: str = "maria",
    ) -> None:
        super().__init__(db_config_name=db_config_name)
        self.table_name = table_name
        self.auto_cast = auto_cast
        self._limit = limit

        self.data_manipulator = GetDataManipulatorMySQL(
            size=size,
            db_config_name=self.db_config_name,
        )

        self.db_utils = DBUtilsMySQL(db_config_name=db_config_name)
        self.table_schema = self.db_utils.get_table_structure(
            self.table_name,
        )
        if db_schema:
            warnings.warn("db_schema is not required")

        if clear_tech_cols:
            warnings.warn("clear_tech_cols is not available")

        self._filters_args: set = set()
        self._select_args: set = set()
        self._filter_operation: str = "and"

    def filter(self, *args, operation: str = "and"):
        self._filters_args: set = set(args)
        self._filter_operation = operation
        return self

    @property
    def _filters(self) -> str:
        filters = "WHERE "
        if len(self._filters_args) == 0:
            filters = "WHERE 1=1"

        filters += f" {self._filter_operation} ".join(self._filters_args)
        return filters

    def select(self, *args):
        self._select_args: set = set()
        self._select_args.update(args)
        return self

    @property
    def _select(self) -> str:
        select = "select "
        if len(self._select_args) == 0:
            select += "*"
        select += ", ".join(self._select_args)
        return select

    @property
    def query(self) -> str:
        limit = ""
        if self._limit is not None:
            limit = f"LIMIT {self._limit}"

        query = f"""
            {self._select}
            FROM {self.table_name}
            {self._filters}
            {limit}
        """
        return query

    @property
    def _clear_statments(self):
        self._filters_args: set = set()
        self._select_args: set = set()
        return self

    def get_df(self, query: Optional[str] = None) -> PandasDataFrame:
        if query is None:
            query = self.query
        df = self.data_manipulator.get_df(query)

        if self.auto_cast:
            df = cast_df_by_schema(df, self.table_schema)
        return df

    def count(self) -> PandasDataFrame:
        query = f"""select count(*) as cnt from ({self.query}) s"""
        return self.get_df(query)["cnt"]
